/**
 * @file
 * VuoWindow implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#import "VuoWindow.h"
#import "VuoWindowApplication.h"
#import "VuoWindowTextInternal.h"
#import "VuoWindowOpenGLInternal.h"

#include "module.h"

#ifdef VUO_COMPILER
VuoModuleDetails({
					 "name" : "VuoWindow",
					 "dependencies" : [
						 "VuoWindowApplication",
						 "VuoWindowTextInternal",
						 "VuoWindowOpenGLInternal"
					 ]
				 });
#endif


/**
 * Creates a VuoWindowApplication instance (if one doesn't already exist).
 *
 * Must be called from the main thread.
 */
void VuoWindowApplicationInit(void)
{
	[NSAutoreleasePool new];
	if (! NSApp)
	{
		// http://stackoverflow.com/a/11010614/238387
		[VuoWindowApplication sharedApplication];
		[NSApp activateIgnoringOtherApps:YES];
		[NSApp run];
	}
}

/**
 * Creates and displays a window containing a text edit widget.
 *
 * Creates a new QApplication instance if one did not already exist.
 *
 * Can be called from any thread but the main thread.
 */
VuoWindowText VuoWindowTextCreate(void)
{
	__block VuoWindowTextInternal *window = NULL;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  VuoWindowApplicationInit();
					  window = [[VuoWindowTextInternal alloc] init];
					  [window makeKeyAndOrderFront:NSApp];
				   });
	return (VuoWindowText *)window;
}

/**
 * Sets up the window to call the trigger functions when its text is edited.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextEnableTriggers
(
		VuoWindowText w,
		VuoOutputTrigger(typedLine, VuoText),
		VuoOutputTrigger(typedWord, VuoText),
		VuoOutputTrigger(typedCharacter, VuoText)
)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window enableTriggersWithTypedLine:typedLine typedWord:typedWord typedCharacter:typedCharacter];
				  });
}

/**
 * Stops the window from calling trigger functions when its text is edited.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextDisableTriggers(VuoWindowText w)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window disableTriggers];
				  });
}

/**
 * Appends text and a linebreak to the text edit widget in the window.
 *
 * Can be called from any thread.
 */
void VuoWindowTextAppendLine(VuoWindowText vw, const char *text)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)vw;
	char *textCopy = strdup(text);  // ... in case the caller frees text before the asynchronous block uses it.
	dispatch_async(dispatch_get_main_queue(), ^{
					   [window appendLine:textCopy];
					   free(textCopy);
				   });
}

/**
 * Closes and deallocates a @c VuoWindowText window.
 *
 * Destroys the current QApplication instance if it no longer contains any widgets.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowTextDestroy(VuoWindowText vw)
{
	VuoWindowTextInternal *window = (VuoWindowTextInternal *)vw;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window close];
					  [window release];
				  });
}


/**
 * Creates and displays a window containing an OpenGL view.
 *
 * Creates a new QApplication instance if one did not already exist.
 *
 * Can be called from any thread but the main thread.
 */
VuoWindowOpenGL VuoWindowOpenGLCreate
(
		void (*initCallback)(void *),
		void (*resizeCallback)(void *, unsigned int, unsigned int),
		void (*drawCallback)(void *),
		void *context
)
{
	__block VuoWindowOpenGLInternal *window = NULL;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  VuoWindowApplicationInit();
					  window = [[VuoWindowOpenGLInternal alloc] initWithInitCallback:initCallback resizeCallback:resizeCallback drawCallback:drawCallback drawContext:context];
					  [window makeKeyAndOrderFront:nil];
				  });
	return (VuoWindowOpenGL *)window;
}

/**
 * Sets up the window to call the trigger functions when events occur.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLEnableTriggers
(
		VuoWindowOpenGL w,
		VuoOutputTrigger(requestedFrame, VuoFrameRequest),
		VuoOutputTrigger(mouseMoved, VuoPoint2d),
		VuoOutputTrigger(mouseLeftPressed, VuoPoint2d)
)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window enableTriggersWithRequestedFrame:requestedFrame mouseMoved:mouseMoved mouseLeftPressed:mouseLeftPressed];
				  });
}

/**
 * Stops the window from calling trigger functions when events occur.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLDisableTriggers(VuoWindowOpenGL w)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window disableTriggers];
				  });
}

/**
 * Schedules the specified window to be redrawn.
 *
 * Can be called from any thread.
 */
void VuoWindowOpenGLRedraw(VuoWindowOpenGL w)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_async(dispatch_get_main_queue(), ^{
					   [window scheduleRedraw];
				   });
}

/**
 * Sets the window's current and preferred aspect ratio.
 *
 * If necessary, the window is resized to match the specified aspect ratio.
 * After calling this method, when the user resizes the window, its height is adjusted to match the specified aspect ratio.
 */
void VuoWindowOpenGLSetAspectRatio(VuoWindowOpenGL w, unsigned int pixelsWide, unsigned int pixelsHigh)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)w;
	dispatch_async(dispatch_get_main_queue(), ^{
					   [window setAspectRatioToWidth:pixelsWide height:pixelsHigh];
				   });
}

/**
 * Closes and deallocates a @c VuoWindowOpenGL window.
 *
 * Destroys the current QApplication instance if it no longer contains any widgets.
 *
 * Can be called from any thread but the main thread.
 */
void VuoWindowOpenGLDestroy(VuoWindowOpenGL vw)
{
	VuoWindowOpenGLInternal *window = (VuoWindowOpenGLInternal *)vw;
	dispatch_sync(dispatch_get_main_queue(), ^{
					  [window close];
					  [window release];
				  });
}
