/**
 * @file
 * VuoWindowOpenGLInternal interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoGlContext.h"

#import <AppKit/AppKit.h>
#import <CoreVideo/CoreVideo.h>


@class VuoWindowOpenGLInternal;

/**
 * Helper for @c VuoWindowOpenGLInternal.
 */
@interface VuoWindowOpenGLView : NSOpenGLView
{
	void (*initCallback)(void *);  ///< Initializes the OpenGL context.
	void (*resizeCallback)(void *, unsigned int width, unsigned int height);  ///< Updates the OpenGL context when the view is resized.
	void (*drawCallback)(void *);  ///< Draws onto the OpenGL context.
	void *drawContext;  ///< Argument to pass to callbacks (e.g. node instance data).
}

- (id)initWithFrame:(NSRect)frame
	   initCallback:(void (*)(void *))_initCallback
	 resizeCallback:(void (*)(void *, unsigned int width, unsigned int height))_resizeCallback
	   drawCallback:(void (*)(void *))_drawCallback
		drawContext:(void *)_drawContext;

@property(retain) VuoWindowOpenGLInternal *glWindow;  ///< The parent window; allows the view to access it while full-screen.

@end


/**
 * A graphics window for use by Vuo node classes.
 *
 * The graphics window renders onto an OpenGL context shared with Vuo nodes, so that it can render the
 * scenes created by those nodes. A scene is rendered by calling callbacks provided by the Vuo nodes.
 */
@interface VuoWindowOpenGLInternal : NSWindow <NSWindowDelegate>
{
	BOOL callbacksEnabled;  ///< True if the window can call its trigger callbacks.

	CVDisplayLinkRef displayLink;	///< The display link, created when -enableTriggers... is called.
	struct VuoWindowOpenGLInternal_displayLinkContext *displayLinkContext;	///< Context data for the display link callback, created when the window is initialized (so that it persists through live-coding reloads).

	void (*mouseMoved)(VuoPoint2d);  ///< @todo (https://b33p.net/kosada/node/5459)
	void (*mouseLeftPressed)(VuoPoint2d);  ///< @todo (https://b33p.net/kosada/node/5459)

	BOOL userResizedWindow;  ///< True if the user has manually resized the window.
	BOOL programmaticallyResizingWindow;  ///< True if a programmatic resize of the window is in progress.
}

@property(retain) VuoWindowOpenGLView *glView;  ///< The OpenGL view inside this window.
@property(retain) NSOpenGLContext *windowedGlContext;  ///< The OpenGL context from Vuo's context pool; allows the windw to access it while the view is full-screen.

- (id)initWithInitCallback:(void (*)(void *))initCallback
			resizeCallback:(void (*)(void *, unsigned int width, unsigned int height))resizeCallback
			  drawCallback:(void (*)(void *))drawCallback
			   drawContext:(void *)drawContext;
- (void)enableTriggersWithRequestedFrame:(void (*)(VuoFrameRequest))requestedFrame
							  mouseMoved:(void (*)(VuoPoint2d))mouseMoved
						mouseLeftPressed:(void (*)(VuoPoint2d))mouseLeftPressed;
- (void)disableTriggers;
- (void)scheduleRedraw;
- (void)setAspectRatioToWidth:(unsigned int)pixelsWide height:(unsigned int)pixelsHigh;
- (void)toggleFullScreen;

@end
