/**
 * @file
 * vuo.math.countWithinRange.integer node implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include <stdlib.h>

VuoModuleDetails({
					 "name" : "Count within Range",
					 "description" :
						 "<p>Keeps track of a count that starts at `minimum` and always stays within `minimum` and `maximum`.</p> \
						 <p>An event to `increment` adds the increment to the count. An event to `decrement` subtracts the decrement from the count.</p> \
						 <p>When `wrapMode` is \"wrap\", the count wraps around to `minimum` when it would be greater than `maximum`, and to `maximum` when it would be less than `minimum`. \
						 When `wrapMode` is \"saturate\", the count stays at `minimum` or `maximum` when it would otherwise go outside the range.</p>",
					 "keywords" : [ "count", "range", "clamp", "loop", "iterate", "saturate", "wrap", "modulus", "clock" ],
					 "version" : "1.0.0",
					 "node": {
						 "isInterface" : false
					 }
				 });


VuoReal * nodeInstanceInit()
{
	VuoReal *countState = (VuoReal *) malloc(sizeof(VuoReal));
	VuoRegister(countState, free);
	*countState = 0;
	return countState;
}

void nodeInstanceEvent
(
		VuoInstanceData(VuoReal *) countState,
		VuoInputData(VuoReal, {"default":1}) increment,
		VuoInputConductor(VuoPortAlwaysConducts,increment) incrementEvent,
		VuoInputData(VuoReal, {"default":1}) decrement,
		VuoInputConductor(VuoPortAlwaysConducts,decrement) decrementEvent,
		VuoInputData(VuoReal, {"default":0}) minimum,
		VuoInputData(VuoReal, {"default":10}) maximum,
		VuoInputData(VuoCountWrapMode, {"default":"wrap"}) wrapMode,
		VuoOutputData(VuoReal) count
)
{
	// Adjust initial count to be within range (in case minimum/maximum have changed).
	if(**countState > maximum)
		**countState = maximum;
	if(**countState < minimum)
		**countState = minimum;

	// Increment/decrement.
	if(incrementEvent)
		**countState += increment;
	if(decrementEvent)
		**countState -= decrement;

	// Adjust incremented/decremented count to be within range.
	switch(wrapMode)
	{
		default:
		case VuoCountWrapMode_Wrap:
			if(**countState > maximum)
				**countState = minimum;
			if(**countState < minimum)
				**countState = maximum;
			break;

		case VuoCountWrapMode_Saturate:
			if(**countState > maximum)
				**countState = maximum;
			if(**countState < minimum)
				**countState = minimum;
			break;
	}

	*count = **countState;
}


void nodeInstanceFini(VuoInstanceData(VuoReal *) countState)
{
}
