/**
 * @file
 * vuo.scene.render.window node implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoWindow.h"
#include "VuoSceneRenderer.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <OpenGL/gl.h>
//#import <OpenGL/gl3.h>
/// @todo After we drop 10.6 support, switch back to gl3 and remove the below 4 lines.  See also r15430 for shader changes.
#include <OpenGL/glext.h>

#define DEBUG 0


VuoModuleDetails({
					 "name" : "Render Scene to Window",
					 "description" : "Creates a window, and renders the objects and their descendants to the window.  When the display is ready for the next frame, `frameRequested` is fired.  When the user moves the mouse inside the window, `mouseMoved` is fired with the current position of the mouse, in root scene coordinates.  (For additional mouse functionality, see `vuo.mouse`.)",
					 "keywords" : [ "draw", "opengl", "scenegraph", "graphics", "display", "view" ],
					 "version" : "1.0.0",
					 "dependencies" : [
						 "VuoSceneRenderer",
						 "VuoWindow"
					 ],
					 "node": {
						 "isInterface" : true
					 }
				 });


struct nodeInstanceData
{
	VuoWindowOpenGL *window;
	VuoSceneRenderer *sceneRenderer;
};

void vuo_scene_render_window_init(void *ctx)
{
	struct nodeInstanceData *context = ctx;

	VuoSceneRenderer_prepareContext(context->sceneRenderer);
}

void vuo_scene_render_window_resize(void *ctx, unsigned int width, unsigned int height)
{
	//fprintf(stderr, "vuo_scene_render_window_resize(%d,%d)\n", width, height);
	struct nodeInstanceData *context = ctx;

	VuoSceneRenderer_regenerateProjectionMatrix(context->sceneRenderer, width, height);
}

void vuo_scene_render_window_draw(void *ctx)
{
	//fprintf(stderr, "vuo_scene_render_window_draw\n");
	struct nodeInstanceData *context = ctx;

	glClearColor(0,0,0,1);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	VuoSceneRenderer_draw(context->sceneRenderer);

#if DEBUG
	VuoSceneRenderer_drawElement(context->sceneRenderer, 0, .08f);	// Normals
	VuoSceneRenderer_drawElement(context->sceneRenderer, 1, .08f);	// Tangents
	VuoSceneRenderer_drawElement(context->sceneRenderer, 2, .08f);	// Bitangents
#endif
}

struct nodeInstanceData *nodeInstanceInit(void)
{
	struct nodeInstanceData *context = (struct nodeInstanceData *)calloc(1,sizeof(struct nodeInstanceData));
	context->sceneRenderer = VuoSceneRenderer_make();
	context->window = VuoWindowOpenGLCreate(
				vuo_scene_render_window_init,
				vuo_scene_render_window_resize,
				vuo_scene_render_window_draw,
				(void *)context
			);
	VuoRegister(context, free);
	return context;
}

void nodeInstanceCallbackStart
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoOutputTrigger(requestedFrame, VuoFrameRequest)
)
{
	VuoWindowOpenGLEnableTriggers((*context)->window, requestedFrame, NULL, NULL);
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) context,
		VuoInputData(VuoList_VuoSceneObject) objects
)
{
	VuoSceneObject rootSceneObject = VuoSceneObject_make(NULL, NULL, VuoTransform_makeIdentity(), objects);
	VuoSceneRenderer_setRootSceneObject((*context)->sceneRenderer, rootSceneObject);

	// Schedule a redraw.
	VuoWindowOpenGLRedraw((*context)->window);
}

void nodeInstanceCallbackStop
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDisableTriggers((*context)->window);
}

void nodeInstanceFini
(
		VuoInstanceData(struct nodeInstanceData *) context
)
{
	VuoWindowOpenGLDestroy((*context)->window);
	VuoSceneRenderer_destroy((*context)->sceneRenderer);
}
