/**
 * @file
 * TestVuoCompilerModule interface and implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "TestVuoCompiler.hh"
#include "VuoCompilerModule.hh"

// Be able to use these types in QTest::addColumn()
Q_DECLARE_METATYPE(vector<string>);

/**
 * Tests for using modules (node classes and types).
 */
class TestVuoCompilerModule : public TestVuoCompiler
{
	Q_OBJECT

private slots:

	void initTestCase()
	{
		initCompiler();
	}

	void cleanupTestCase()
	{
		cleanupCompiler();
	}

	void testModuleSymbolsRenamed()
	{
		string testClassID = "vuo.math.add.integer";
		VuoCompilerNodeClass *testClass = compiler->getNodeClass(testClassID);
		Module *module = testClass->getModule();
		string nodeEventFuncName = testClass->nameForGlobal("nodeEvent");
		Function *functionByNameFromNodeClass = module->getFunction(nodeEventFuncName);
		Function *functionByLiteralName = module->getFunction("vuo_math_add_integer__nodeEvent");
		QVERIFY(functionByNameFromNodeClass != NULL);
		QVERIFY(functionByNameFromNodeClass == functionByLiteralName);
	}

	void testDependencies_data()
	{
		QTest::addColumn< QString >("nodeClass");
		QTest::addColumn< vector<string> >("expectedDependencies");

		{
			vector<string> dependencies;
			QTest::newRow("Add does not have dependencies.") << "vuo.math.add.integer" << dependencies;
		}

		{
			vector<string> dependencies;
			dependencies.push_back("c");
			QTest::newRow("Append has dependencies.") << "vuo.text.append" << dependencies;
		}
	}
	void testDependencies()
	{
		QFETCH(QString, nodeClass);
		QFETCH(vector<string>, expectedDependencies);

		VuoCompilerNodeClass *cnc = compiler->getNodeClass( qPrintable(nodeClass) );
		vector<string> actualDependencies = cnc->getDependencies();

		QCOMPARE(actualDependencies.size(), expectedDependencies.size());
		for (uint i = 0; i < actualDependencies.size(); ++i)
		{
			string actualDependency = actualDependencies.at(i);
			string expectedDependency = expectedDependencies.at(i);
			QCOMPARE(QString(actualDependency.c_str()), QString(expectedDependency.c_str()));
		}
	}

	void testKeywords_data()
	{
		QTest::addColumn< QString >("nodeClass");
		QTest::addColumn< vector<string> >("expectedKeywords");

		{
			vector<string> keywords;
			QTest::newRow("'Are Equal' does not have keywords.") << "vuo.text.areEqual" << keywords;
		}

		{
			vector<string> keywords;
			keywords.push_back("length");
			keywords.push_back("size");
			QTest::newRow("Count Characters has keywords.") << "vuo.text.countCharacters" << keywords;
		}
	}
	void testKeywords()
	{
		QFETCH(QString, nodeClass);
		QFETCH(vector<string>, expectedKeywords);

		VuoCompilerNodeClass *cnc = compiler->getNodeClass( qPrintable(nodeClass) );
		vector<string> actualKeywords = cnc->getBase()->getKeywords();

		QCOMPARE(actualKeywords.size(), expectedKeywords.size());
		for (uint i = 0; i < actualKeywords.size(); ++i)
		{
			string actualKeyword = actualKeywords.at(i);
			string expectedKeyword = expectedKeywords.at(i);
			QCOMPARE(QString(actualKeyword.c_str()), QString(expectedKeyword.c_str()));
		}
	}
};

QTEST_APPLESS_MAIN(TestVuoCompilerModule)
#include "TestVuoCompilerModule.moc"
